using System;
using System.Collections.Generic;
using System.Text;
using System.Windows.Threading;
using System.ComponentModel;
using System.Collections;
using System.Collections.Specialized;
using System.Globalization;
using Wintellect.PowerCollections;
using System.Collections.ObjectModel;
using System.Diagnostics;

namespace CollectionViewInfragistics
{
    public class AutoFilterCollectionView<T> : DispatcherObject, ICollectionView, IEnumerable, INotifyCollectionChanged, INotifyPropertyChanged
        where T : INotifyPropertyChanged
    {
        private static readonly CurrentChangingEventArgs _currentChangingEventArgs = new CurrentChangingEventArgs(false);

        private int _deferLevel;
        private int _currentIndex;
        private object _currentItem;

        private CultureInfo _cultureInfo;

        private Dictionary<string, FilterInfo<T>> _filters = new Dictionary<string, FilterInfo<T>>();
        private IList<T> _collection;
        private Dictionary<T, ItemInfo> _linkDictionary = new Dictionary<T, ItemInfo>();
        private OrderedSet<ItemInfo> _localCollection = new OrderedSet<ItemInfo>();

        public AutoFilterCollectionView(IList<T> collection)
        {
            if (null == collection)
            {
                throw new ArgumentNullException("collection");
            }
            _collection = collection;

            INotifyCollectionChanged collectionChanged = collection as INotifyCollectionChanged;
            if (collectionChanged != null)
            {
                collectionChanged.CollectionChanged += new NotifyCollectionChangedEventHandler(OnCollectionChanged);
            }

            BuildLocalCollection();
        }

        private void BuildLocalCollection()
        {
            _localCollection.Clear();
            _linkDictionary.Clear();
            for (int i = 0; i < _collection.Count; ++i)
            {
                T item = _collection[i];
                if (PassesFilter(item))
                {
                    ItemInfo itemInfo = new ItemInfo(i, true);
                    _localCollection.Add(itemInfo);
                    
                    _linkDictionary[item] = itemInfo;// index;
                }
                item.PropertyChanged += new PropertyChangedEventHandler(OnItemPropertyChanged);
            }
        }

        void OnItemPropertyChanged(object sender, PropertyChangedEventArgs e)
        {
            T item = (T)sender;
            if (_filters.ContainsKey(e.PropertyName))
            {
                if (PassesFilter(item))
                {
                    AddItem(item);
                }
                else
                {
                    RemoveItem(item);
                }
            }
        }

        protected void SetCurrent(object newItem, int newPosition)
        {
            _currentItem = newItem;
            _currentIndex = newPosition;
        }

        public virtual int IndexOf(T item)
        {
            if (_linkDictionary.ContainsKey(item))
            {
                return _localCollection.IndexOf(_linkDictionary[item]);
            }
            return -1;
        }

        void OnCollectionChanged(object sender, NotifyCollectionChangedEventArgs e)
        {
            switch (e.Action)
            {
                case NotifyCollectionChangedAction.Add:
                    foreach (T item in e.NewItems)
                    {
                        AddItem(item);
                        item.PropertyChanged += new PropertyChangedEventHandler(OnItemPropertyChanged);
                    }
                    break;

                case NotifyCollectionChangedAction.Remove:
                    foreach (T item in e.OldItems)
                    {
                        RemoveItem(item);
                        item.PropertyChanged -= new PropertyChangedEventHandler(OnItemPropertyChanged);
                    }
                    break;

                case NotifyCollectionChangedAction.Replace:
                    foreach (T item in e.NewItems)
                    {
                        AddItem(item);
                        item.PropertyChanged += new PropertyChangedEventHandler(OnItemPropertyChanged);
                    }

                    foreach (T item in e.OldItems)
                    {
                        RemoveItem(item);
                        item.PropertyChanged -= new PropertyChangedEventHandler(OnItemPropertyChanged);
                    }
                    break;

                case NotifyCollectionChangedAction.Reset:
                    _localCollection.Clear();
                    _linkDictionary.Clear();
                    OnCollectionChanged(new NotifyCollectionChangedEventArgs(NotifyCollectionChangedAction.Reset));
                    break;
            }
        }

        private bool RemoveItem(T item)
        {
            if (_linkDictionary.ContainsKey(item))
            {
                ItemInfo itemInfo = _linkDictionary[item];
                int index = _localCollection.IndexOf(itemInfo);
                _localCollection.Remove(_linkDictionary[item]);
                _linkDictionary.Remove(item);

                Debug.WriteLine(string.Format("Remove item event. Item index{0}", index));
                OnCollectionChanged(new NotifyCollectionChangedEventArgs(NotifyCollectionChangedAction.Remove, item, index));

                return true;
            }

            return false;
        }

        private bool AddItem(T item)
        {
            if (!PassesFilter(item))
            {
                return false;
            }
            bool alreadyAdded = true;
            ItemInfo itemInfo;
            if (_linkDictionary.ContainsKey(item))
            {
                itemInfo = _linkDictionary[item];
                itemInfo.PassesFilter = true;// PassesFilter(item);
            }
            else
            {
                alreadyAdded = false;
                itemInfo = new ItemInfo();
                itemInfo.OrginalCollectionIndex = _collection.IndexOf(item);
                itemInfo.PassesFilter = true;// PassesFilter(item);
                _localCollection.Add(itemInfo);
                _linkDictionary[item] = itemInfo;
            }

            int index = _localCollection.IndexOf(itemInfo);

            if (!alreadyAdded)
            {
                Debug.WriteLine(string.Format("Add item event. Item index{0}", index));
                OnCollectionChanged(new NotifyCollectionChangedEventArgs(NotifyCollectionChangedAction.Add, item, index));
            }
            return alreadyAdded;
        }

        private void OnCollectionChanged(NotifyCollectionChangedEventArgs args)
        {
            if (args == null)
            {
                throw new ArgumentNullException("args");
            }
            if (this.CollectionChanged != null)
            {
                this.CollectionChanged(this, args);
            }
            if ((args.Action != NotifyCollectionChangedAction.Replace) && (args.Action != NotifyCollectionChangedAction.Move))
            {
                this.OnPropertyChanged("Count");
            }
        }

        public virtual bool PassesFilter(T item)
        {
            bool result = true;
            if (CanFilter && (_filters.Count > 0))
            {
                foreach (FilterInfo<T> filter in _filters.Values)
                {
                    if (!filter.IsMatch(item))
                    {
                        result = false;
                        break;
                    }
                }
            }
            return result;
        }

        #region ICollectionView Members

        public bool CanFilter
        {
            get
            {
                return true;
            }
        }

        public bool CanGroup
        {
            get
            {
                return false;
            }
        }

        public bool CanSort
        {
            get
            {
                return false;
            }
        }

        public bool Contains(object item)
        {
            if (!(item is T))
            {
                return false;
            }
            return _linkDictionary.ContainsKey((T)item);
        }

        public CultureInfo Culture
        {
            get
            {
                return _cultureInfo;
            }
            set
            {
                if (null == value)
                {
                    throw new ArgumentNullException("Culture");
                }
                if (value != _cultureInfo)
                {
                    _cultureInfo = value;
                    OnPropertyChanged("Culture");
                }
            }
        }

        public event EventHandler CurrentChanged;

        protected virtual void OnCurrentChanged()
        {
            if (this.CurrentChanged != null)
            {
                this.CurrentChanged(this, EventArgs.Empty);
            }
        }

        public event CurrentChangingEventHandler CurrentChanging;

        protected void OnCurrentChanging()
        {
            _currentIndex = -1;
            OnCurrentChanging(_currentChangingEventArgs);
        }

        protected virtual void OnCurrentChanging(CurrentChangingEventArgs args)
        {
            if (args == null)
            {
                throw new ArgumentNullException("args");
            }
            if (this.CurrentChanging != null)
            {
                this.CurrentChanging(this, args);
            }
        }


        public object CurrentItem
        {
            get
            {
                Debug.WriteLine("AutoFilterCollectionView.CurrentItem");
                return _currentItem;
            }
        }

        public int CurrentPosition
        {
            get
            {
                Debug.WriteLine("AutoFilterCollectionView.CurrentPosition");
                return _currentIndex;
            }
        }

        public IDisposable DeferRefresh()
        {
            Debug.WriteLine("AutoFilterCollectionView.DeferRefresh");
            _deferLevel++;
            return new DeferHelper(this);
        }

        private void EndDefer()
        {
            Debug.WriteLine("AutoFilterCollectionView.EndDefer");
            this._deferLevel--;
            if (this._deferLevel == 0)
            {
                this.Refresh();
            }
        }

        Predicate<object> ICollectionView.Filter
        {
            get
            {
                return null;
                // return _filter;
            }
            set
            {
                //if (!CanFilter)
                //{
                //    throw new NotSupportedException();
                //}

                //_filter = value;
                //Refresh();
            }
        }

        public void AddPropertyFilter(PropertyDescriptor propertyDescriptor, object value)
        {
            if (null == propertyDescriptor)
            {
                throw new ArgumentNullException("propertyDescriptor");
            }
            if (!_filters.ContainsKey(propertyDescriptor.Name))
            {
                FilterInfo<T> newFilter = new FilterInfo<T>(propertyDescriptor, value);
                _filters[propertyDescriptor.Name] = newFilter;
                Refresh();
            }
            else
            {
                object existingValue = _filters[propertyDescriptor.Name].Value;
                if ((existingValue == null && value != null) || !existingValue.Equals(value))
                {
                    _filters[propertyDescriptor.Name].Value = value;
                    Refresh();
                }
            }
        }

        public void RemovePropertyFilter(PropertyDescriptor propertyDescriptor)
        {
            if (null == propertyDescriptor)
            {
                throw new ArgumentNullException("propertyDescriptor");
            }
            if (_filters.ContainsKey(propertyDescriptor.Name))
            {
                _filters.Remove(propertyDescriptor.Name);
                Refresh();
            }
        }

        public virtual ObservableCollection<GroupDescription> GroupDescriptions
        {
            get
            {
                return null;
            }
        }

        public ReadOnlyObservableCollection<object> Groups
        {
            get
            {
                return null;
            }
        }

        public int Count
        {
            get
            {
                Debug.WriteLine("AutoFilterCollectionView.Count");
                return _localCollection.Count;
            }
        }

        public bool IsCurrentAfterLast
        {
            get
            {
                Debug.WriteLine("AutoFilterCollectionView.IsCurrentAfterLast");
                return _currentIndex >= Count;
            }
        }

        public bool IsCurrentBeforeFirst
        {
            get
            {
                Debug.WriteLine("AutoFilterCollectionView.IsCurrentBeforeFirst");
                if (!IsEmpty)
                {
                    return (_currentIndex < 0);
                }
                return true;
            }
        }

        public bool IsEmpty
        {
            get
            {

                return _localCollection.Count == 0;
            }
        }

        private void MoveCurrentToPositionInternal(int index)
        {
            if (index < 0)
            {
                this.SetCurrent(null, -1);
            }
            else if (index >= Count)
            {
                this.SetCurrent(null, Count);
            }
            else
            {
                this.SetCurrent(_collection[_localCollection[index].OrginalCollectionIndex], index);
            }
        }

        public bool MoveCurrentTo(object item)
        {
            Debug.WriteLine("AutoFilterCollectionView.MoveCurrentTo");
            T element = (T)item;
            if (object.Equals(CurrentItem, item) && ((item != null) || this.IsCurrentInView))
            {
                return this.IsCurrentInView;
            }
            int position = -1;
            if (PassesFilter(element))
            {
                AddItem(element);
                position = IndexOf(element);
            }
            else
            {
                RemoveItem(element);
            }
            return this.MoveCurrentToPosition(position);
        }

        public bool MoveCurrentToFirst()
        {
            Debug.WriteLine("AutoFilterCollectionView.MoveCurrentToFirst");
            return MoveCurrentToPosition(0);
        }

        public bool MoveCurrentToLast()
        {
            Debug.WriteLine("AutoFilterCollectionView.MoveCurrentToLast");
            return MoveCurrentToPosition(Count - 1);
        }

        public bool MoveCurrentToNext()
        {
            Debug.WriteLine("AutoFilterCollectionView.MoveCurrentToNext");
            return ((CurrentPosition < Count) && MoveCurrentToPosition(CurrentPosition + 1));
        }

        public bool MoveCurrentToPosition(int position)
        {
            Debug.WriteLine("AutoFilterCollectionView.MoveCurrentToPosition");
            if ((position < -1) || (position > Count))
            {
                throw new ArgumentOutOfRangeException("position");
            }
            if (((position != CurrentPosition) || !IsCurrentInSync) && CanCurrentChange())
            {
                bool isCurrentAfterLast = IsCurrentAfterLast;
                bool isCurrentBeforeFirst = IsCurrentBeforeFirst;
                MoveCurrentToPositionInternal(position);
                this.OnCurrentChanged();
                if (this.IsCurrentAfterLast != isCurrentAfterLast)
                {
                    this.OnPropertyChanged("IsCurrentAfterLast");
                }
                if (this.IsCurrentBeforeFirst != isCurrentBeforeFirst)
                {
                    this.OnPropertyChanged("IsCurrentBeforeFirst");
                }
                this.OnPropertyChanged("CurrentPosition");
                this.OnPropertyChanged("CurrentItem");
            }
            return this.IsCurrentInView;
        }

        public bool MoveCurrentToPrevious()
        {
            Debug.WriteLine("AutoFilterCollectionView.MoveCurrentToPrevious");
            return ((CurrentPosition >= 0) && MoveCurrentToPosition(CurrentPosition - 1));
        }

        public void Refresh()
        {
            Debug.WriteLine("AutoFilterCollectionView.Refresh");
            if (SortDescriptions.Count > 0)
            {
                throw new InvalidOperationException("Can not refresh Refresh");
            }

            bool oldIsCurrentAfterLast = IsCurrentAfterLast;
            bool oldIsCurrentBeforeFirst = IsCurrentBeforeFirst;
            object oldItem = _currentItem;
            int oldIndex = _currentIndex;

            BuildLocalCollection();
            OnCurrentChanging();
            if (this.IsEmpty || IsCurrentBeforeFirst)
            {
                MoveCurrentToPositionInternal(-1);
            }
            else if (IsCurrentAfterLast)
            {
                MoveCurrentToPositionInternal(Count);
            }
            else if (_currentItem != null)
            {
                T item = (T)_currentItem;
                int index = 0;
                if (_linkDictionary.ContainsKey(item))
                {
                    index = _localCollection.IndexOf(_linkDictionary[item]);
                }
                MoveCurrentToPositionInternal(index);
            }

            this.OnCollectionChanged(new NotifyCollectionChangedEventArgs(NotifyCollectionChangedAction.Reset));
            this.OnCurrentChanged();
            if (this.IsCurrentAfterLast != oldIsCurrentAfterLast)
            {
                this.OnPropertyChanged("IsCurrentAfterLast");
            }
            if (this.IsCurrentBeforeFirst != oldIsCurrentBeforeFirst)
            {
                this.OnPropertyChanged("IsCurrentBeforeFirst");
            }
            if (oldIndex != _currentIndex)
            {
                this.OnPropertyChanged("CurrentPosition");
            }
            if (oldItem != _currentItem)
            {
                this.OnPropertyChanged("CurrentItem");
            }
        }

        protected bool CanCurrentChange()
        {
            CurrentChangingEventArgs args = new CurrentChangingEventArgs();
            OnCurrentChanging(args);
            return !args.Cancel;
        }

        protected bool IsCurrentInSync
        {
            get
            {
                if (this.IsCurrentInView)
                {
                    return (((object)_collection[_localCollection[CurrentPosition].OrginalCollectionIndex]) == _currentItem);
                }
                return (this.CurrentItem == null);
            }
        }

        private bool IsCurrentInView
        {
            get
            {
                return ((0 <= _currentIndex) && (_currentIndex < Count));
            }
        }

        public virtual SortDescriptionCollection SortDescriptions
        {
            get
            {
                return SortDescriptionCollection.Empty;
            }
        }

        public IEnumerable SourceCollection
        {
            get
            {
                return _collection;
            }
        }

        #endregion

        #region IEnumerable Members

        public IEnumerator GetEnumerator()
        {
            Debug.WriteLine("AutoFilterCollectionView.GetEnumerator");
            foreach (ItemInfo itemInfo in _localCollection)
            {
                yield return _collection[itemInfo.OrginalCollectionIndex];
            }
        }

        #endregion

        #region INotifyCollectionChanged Members

        public event NotifyCollectionChangedEventHandler CollectionChanged;

        #endregion

        #region Nested Class

        private class ItemInfo : IComparable<ItemInfo>
        {
            private int _orginalCollectionIndex;
            private bool _passesFilter;

            public ItemInfo()
            { }

            public ItemInfo(int originalCollectionIndex, bool passesFilter)
            {
                Debug.Assert(originalCollectionIndex >= 0);
                _orginalCollectionIndex = originalCollectionIndex;
                _passesFilter = passesFilter;
            }

            public int OrginalCollectionIndex
            {
                get { return _orginalCollectionIndex; }
                set { _orginalCollectionIndex = value; }
            }

            public bool PassesFilter
            {
                get { return _passesFilter; }
                set { _passesFilter = value; }
            }

            #region IComparable Members

            public int CompareTo(ItemInfo obj)
            {
                return _orginalCollectionIndex.CompareTo(obj._orginalCollectionIndex);
            }

            #endregion
        }


        public class DeferHelper : IDisposable
        {
            private AutoFilterCollectionView<T> _collectionView;

            public DeferHelper(AutoFilterCollectionView<T> collectionView)
            {
                _collectionView = collectionView;
            }

            #region IDisposable Members

            public void Dispose()
            {
                if (null != _collectionView)
                {
                    _collectionView.EndDefer();
                    _collectionView = null;
                }
            }

            #endregion
        }

        private class FilterInfo<T> 
        {
            private PropertyDescriptor _propertyDescriptor;
            private object _value;

            public FilterInfo(PropertyDescriptor propertyDescriptor, object value)
            {
                _propertyDescriptor = propertyDescriptor;
                _value = value;
            }

            public string PropertyName
            {
                get
                {
                    return _propertyDescriptor.Name;
                }
            }

            public PropertyDescriptor PropertyDescriptor
            {
                get
                {
                    return _propertyDescriptor;
                }
            }

            public object Value
            {
                get
                {
                    return _value;
                }
                set
                {
                    _value = value;
                }
            }

            public bool IsMatch(T item)
            {
                object value = _propertyDescriptor.GetValue(item);
                if (value == null)
                {
                    if (_value == null)
                    {
                        return true;
                    }
                    return false;
                }

                return value.Equals(_value);
            }
        }
        #endregion

        #region INotifyPropertyChanged Members

        public event PropertyChangedEventHandler PropertyChanged;

        protected virtual void OnPropertyChanged(string propertyName)
        {
            if (string.IsNullOrEmpty(propertyName))
            {
                throw new ArgumentException("propertyName");
            }
            if (this.PropertyChanged != null)
            {
                this.PropertyChanged(this, new PropertyChangedEventArgs(propertyName));
            }
        }

        #endregion
    }
}
